@@ -8,19 +8,16 @@ import io
 import os
 import sys
 import tokenize
-from typing import Any
-from typing import Callable
 from typing import Generator
 from typing import Iterable
-from typing import List
 from typing import NamedTuple
 from typing import Sequence
-from typing import Tuple
 
-from aspy.refactor_imports.import_obj import AbstractImportObj
-from aspy.refactor_imports.import_obj import import_obj_from_str
-from aspy.refactor_imports.import_obj import ImportImport
-from aspy.refactor_imports.sort import sort
+from classify_imports import Import
+from classify_imports import import_obj_from_str
+from classify_imports import ImportFrom
+from classify_imports import Settings
+from classify_imports import sort
 
 # this is a performance hack.  see https://bugs.python.org/issue43014
 if (  # pragma: no branch
@@ -32,6 +29,8 @@ if (  # pragma: no branch
 
 CodeType = enum.Enum('CodeType', 'PRE_IMPORT_CODE IMPORT NON_CODE CODE')
 
+NON_COMBINABLE = (CodeType.IMPORT, CodeType.PRE_IMPORT_CODE)
+
 
 class CodePartition(NamedTuple):
     code_type: CodeType
@@ -42,39 +41,16 @@ TERMINATES_COMMENT = frozenset((tokenize.NL, tokenize.ENDMARKER))
 TERMINATES_DOCSTRING = frozenset((tokenize.NEWLINE, tokenize.ENDMARKER))
 TERMINATES_IMPORT = frozenset((tokenize.NEWLINE, tokenize.ENDMARKER))
 
-ImportToReplace = Tuple[List[str], List[str], str]
-Step = Callable[[Any], List[CodePartition]]
-
-
-def _src_to_import_lines(src: str) -> set[int]:
-    return {
-        node.lineno for node in ast.parse(src).body
-        if isinstance(node, (ast.Import, ast.ImportFrom))
-    }
-
-
-def get_line_offsets_by_line_no(src: str) -> list[int]:
-    # Padded so we can index with line number
-    offsets = [0, 0]
-    for line in src.splitlines(True):
-        offsets.append(offsets[-1] + len(line))
-    return offsets
-
 
 def _partitions_to_src(partitions: Iterable[CodePartition]) -> str:
     return ''.join(part.src for part in partitions)
 
 
 def partition_source(src: str) -> list[CodePartition]:
-    """Partitions source into a list of `CodePartition`s for import
-    refactoring.
-    """
-    top_level_import_line_numbers = _src_to_import_lines(src)
-
-    line_offsets = get_line_offsets_by_line_no(src)
+    lines = src.splitlines(True)
 
     chunks = []
-    startpos = 0
+    startline = 0
     pending_chunk_type = None
     possible_ending_tokens = None
     seen_import = False
@@ -85,7 +61,8 @@ def partition_source(src: str) -> list[CodePartition]:
         if pending_chunk_type is None:
             if not seen_import and token_type == tokenize.COMMENT:
                 if 'noreorder' in token_text:
-                    chunks.append(CodePartition(CodeType.CODE, src[startpos:]))
+                    srctext = ''.join(lines[startline:])
+                    chunks.append(CodePartition(CodeType.CODE, srctext))
                     break
                 else:
                     pending_chunk_type = CodeType.PRE_IMPORT_CODE
@@ -93,20 +70,24 @@ def partition_source(src: str) -> list[CodePartition]:
             elif not seen_import and token_type == tokenize.STRING:
                 pending_chunk_type = CodeType.PRE_IMPORT_CODE
                 possible_ending_tokens = TERMINATES_DOCSTRING
-            elif scol == 0 and srow in top_level_import_line_numbers:
+            elif (
+                    scol == 0 and
+                    token_type == tokenize.NAME and
+                    token_text in {'from', 'import'}
+            ):
                 seen_import = True
                 pending_chunk_type = CodeType.IMPORT
                 possible_ending_tokens = TERMINATES_IMPORT
             elif token_type == tokenize.NL:
                 # A NL token is a non-important newline, we'll immediately
                 # append a NON_CODE partition
-                endpos = line_offsets[erow] + ecol
-                srctext = src[startpos:endpos]
-                startpos = endpos
+                srctext = ''.join(lines[startline:erow])
+                startline = erow
                 chunks.append(CodePartition(CodeType.NON_CODE, srctext))
             elif token_type == tokenize.COMMENT:
                 if 'noreorder' in token_text:
-                    chunks.append(CodePartition(CodeType.CODE, src[startpos:]))
+                    srctext = ''.join(lines[startline:])
+                    chunks.append(CodePartition(CodeType.CODE, srctext))
                     break
                 else:
                     pending_chunk_type = CodeType.CODE
@@ -118,40 +99,31 @@ def partition_source(src: str) -> list[CodePartition]:
                 # We've reached a `CODE` block, which spans the rest of the
                 # file (intentionally timid).  Let's append that block and be
                 # done
-                chunks.append(CodePartition(CodeType.CODE, src[startpos:]))
+                srctext = ''.join(lines[startline:])
+                chunks.append(CodePartition(CodeType.CODE, srctext))
                 break
         # Attempt to find ending of token
         elif token_type in possible_ending_tokens:
-            endpos = line_offsets[erow] + ecol
-            srctext = src[startpos:endpos]
-            startpos = endpos
+            srctext = ''.join(lines[startline:erow])
+            startline = erow
             chunks.append(CodePartition(pending_chunk_type, srctext))
             pending_chunk_type = None
             possible_ending_tokens = None
         elif token_type == tokenize.COMMENT and 'noreorder' in token_text:
-            chunks.append(CodePartition(CodeType.CODE, src[startpos:]))
+            srctext = ''.join(lines[startline:])
+            chunks.append(CodePartition(CodeType.CODE, srctext))
             break
 
-    chunks = [chunk for chunk in chunks if chunk.src]
+    # combine any trailing code chunks
+    ret = [chunk for chunk in chunks if chunk.src]
+    code = []
+    while ret and ret[-1].code_type not in NON_COMBINABLE:
+        code.append(ret.pop())
+    if code:
+        code_src = _partitions_to_src(reversed(code))
+        ret.append(CodePartition(CodeType.CODE, code_src))
 
-    # Make sure we're not removing any code
-    assert _partitions_to_src(chunks) == src
-    return chunks
-
-
-def combine_trailing_code_chunks(
-        partitions: Iterable[CodePartition],
-) -> list[CodePartition]:
-    chunks = list(partitions)
-
-    NON_COMBINABLE = (CodeType.IMPORT, CodeType.PRE_IMPORT_CODE)
-    if chunks and chunks[-1].code_type not in NON_COMBINABLE:
-        src = chunks.pop().src
-        while chunks and chunks[-1].code_type not in NON_COMBINABLE:
-            src = chunks.pop().src + src
-
-        chunks.append(CodePartition(CodeType.CODE, src))
-    return chunks
+    return ret
 
 
 def separate_comma_imports(
@@ -161,12 +133,10 @@ def separate_comma_imports(
     def _inner() -> Generator[CodePartition, None, None]:
         for partition in partitions:
             if partition.code_type is CodeType.IMPORT:
-                import_obj = import_obj_from_str(partition.src)
-                if import_obj.has_multiple_imports:
-                    for new_import_obj in import_obj.split_imports():
-                        yield CodePartition(
-                            CodeType.IMPORT, new_import_obj.to_text(),
-                        )
+                obj = import_obj_from_str(partition.src)
+                if obj.is_multiple:
+                    for new_obj in obj.split():
+                        yield CodePartition(CodeType.IMPORT, str(new_obj))
                 else:
                     yield partition
             else:
@@ -176,49 +146,49 @@ def separate_comma_imports(
 
 
 def add_imports(
-        partitions: Iterable[CodePartition],
+        partitions: list[CodePartition],
         to_add: tuple[str, ...] = (),
 ) -> list[CodePartition]:
-    partitions = list(partitions)
-    if not _partitions_to_src(partitions).strip():
-        return partitions
-
-    # If we don't have a trailing newline, this refactor is wrong
-    if not partitions[-1].src.endswith('\n'):
-        partitions[-1] = partitions[-1]._replace(src=partitions[-1].src + '\n')
-
     return partitions + [
         CodePartition(CodeType.IMPORT, imp_statement.strip() + '\n')
         for imp_statement in to_add
     ]
 
 
-def remove_imports(
-        partitions: Iterable[CodePartition],
-        to_remove: tuple[str, ...] = (),
-) -> list[CodePartition]:
-    to_remove_imports: set[AbstractImportObj] = set()
-    for s in to_remove:
-        to_remove_imports.update(import_obj_from_str(s).split_imports())
+class Replacements(NamedTuple):
+    # (orig_mod, attr) => new_mod
+    exact: dict[tuple[str, str], str]
+    # orig_mod => new_mod (no attr)
+    mods: dict[str, str]
 
-    def _inner() -> Generator[CodePartition, None, None]:
-        for partition in partitions:
-            if (
-                    partition.code_type is not CodeType.IMPORT or
-                    import_obj_from_str(partition.src) not in to_remove_imports
-            ):
-                yield partition
+    @classmethod
+    def make(cls, args: list[tuple[str, str, str]]) -> Replacements:
+        exact = {}
+        mods = {}
 
-    return list(_inner())
+        for mod_from, mod_to, attr in args:
+            if attr:
+                exact[mod_from, attr] = mod_to
+            else:
+                mod_from_base, _, mod_from_attr = mod_from.rpartition('.')
+                mod_to_base, _, mod_to_attr = mod_to.rpartition('.')
 
+                # for example `six.moves.urllib.request=urllib.request`
+                if (
+                        mod_from_attr and
+                        mod_to_base and
+                        mod_from_attr == mod_to_attr
+                ):
+                    exact[mod_from_base, mod_from_attr] = mod_to_base
 
-def _mod_startswith(mod_parts: list[str], prefix_parts: list[str]) -> bool:
-    return mod_parts[:len(prefix_parts)] == prefix_parts
+                mods[mod_from] = mod_to
+
+        return cls(exact=exact, mods=mods)
 
 
 def replace_imports(
         partitions: Iterable[CodePartition],
-        to_replace: Iterable[ImportToReplace] = (),
+        to_replace: Replacements,
 ) -> list[CodePartition]:
     def _inner() -> Generator[CodePartition, None, None]:
         for partition in partitions:
@@ -226,51 +196,63 @@ def replace_imports(
                 import_obj = import_obj_from_str(partition.src)
 
                 # cannot rewrite import-imports: makes undefined names
-                if isinstance(import_obj, ImportImport):
+                if isinstance(import_obj, Import):
                     yield partition
                     continue
 
-                mod_parts = import_obj.import_statement.module.split('.')
-                symbol = import_obj.import_statement.symbol
-                asname = import_obj.import_statement.asname
+                mod, symbol, asname = import_obj.key
+                mod_symbol = f'{mod}.{symbol}'
 
-                for orig_mod, new_mod, attr in to_replace:
-                    if (
-                            (attr == symbol and mod_parts == orig_mod) or
-                            (not attr and _mod_startswith(mod_parts, orig_mod))
-                    ):
-                        mod_parts[:len(orig_mod)] = new_mod
-                        import_obj.ast_obj.module = '.'.join(mod_parts)
-                        new_src = import_obj.to_text()
-                        yield partition._replace(src=new_src)
-                        break
-                    # from a.b.c import d => from c import d
-                    elif (
-                            (mod_parts + [symbol] == orig_mod) and
-                            not attr and
-                            len(new_mod) > 1 and
-                            (asname or symbol == new_mod[-1])
-                    ):
-                        import_obj.ast_obj.module = '.'.join(new_mod[:-1])
-                        import_obj.ast_obj.names = [
-                            ast.alias(name=new_mod[-1], asname=asname),
-                        ]
-                        new_src = import_obj.to_text()
-                        yield partition._replace(src=new_src)
-                        break
-                    # from x.y import z => import z
-                    elif (
-                            not attr and
-                            mod_parts + [symbol] == orig_mod and
-                            len(new_mod) == 1
-                    ):
-                        mod_name, = new_mod
-                        asname_src = f' as {asname}' if asname else ''
-                        new_src = f'import {mod_name}{asname_src}\n'
-                        yield partition._replace(src=new_src)
-                        break
+                # from a.b.c import d => from e.f.g import d
+                if (mod, symbol) in to_replace.exact:
+                    node = ast.ImportFrom(
+                        module=to_replace.exact[mod, symbol],
+                        names=import_obj.node.names,
+                        level=0,
+                    )
+                    yield partition._replace(src=str(ImportFrom(node)))
+                # from a.b.c import d as e => from f import g as e
+                # from a.b.c import d as e => import f as e
+                # from a.b import c => import c
+                elif (
+                        mod_symbol in to_replace.mods and
+                        (asname or to_replace.mods[mod_symbol] == symbol)
+                ):
+                    new_mod = to_replace.mods[mod_symbol]
+                    new_mod, dot, new_sym = new_mod.rpartition('.')
+                    if new_mod:
+                        node = ast.ImportFrom(
+                            module=new_mod,
+                            names=[ast.alias(new_sym, asname)],
+                            level=0,
+                        )
+                        yield partition._replace(src=str(ImportFrom(node)))
+                    elif not dot:
+                        node_i = ast.Import(names=[ast.alias(new_sym, asname)])
+                        yield partition._replace(src=str(Import(node_i)))
+                    else:
+                        yield partition
+                # from a.b.c import d => from e import d
+                elif mod in to_replace.mods:
+                    node = ast.ImportFrom(
+                        module=to_replace.mods[mod],
+                        names=import_obj.node.names,
+                        level=0,
+                    )
+                    yield partition._replace(src=str(ImportFrom(node)))
                 else:
-                    yield partition
+                    for mod_name in _module_to_base_modules(mod):
+                        if mod_name in to_replace.mods:
+                            new_mod = to_replace.mods[mod_name]
+                            node = ast.ImportFrom(
+                                module=f'{new_mod}{mod[len(mod_name):]}',
+                                names=import_obj.node.names,
+                                level=0,
+                            )
+                            yield partition._replace(src=str(ImportFrom(node)))
+                            break
+                    else:
+                        yield partition
             else:
                 yield partition
     return list(_inner())
@@ -287,24 +269,24 @@ def _module_to_base_modules(s: str) -> Generator[str, None, None]:
 
 def remove_duplicated_imports(
         partitions: Iterable[CodePartition],
+        *,
+        to_remove: set[tuple[str, ...]],
 ) -> list[CodePartition]:
-    seen: set[AbstractImportObj] = set()
+    seen = set(to_remove)
     seen_module_names: set[str] = set()
     without_exact_duplicates = []
 
     for partition in partitions:
         if partition.code_type is CodeType.IMPORT:
             import_obj = import_obj_from_str(partition.src)
-            if import_obj not in seen:
-                seen.add(import_obj)
+            if import_obj.key not in seen:
+                seen.add(import_obj.key)
                 if (
-                        isinstance(import_obj, ImportImport) and
-                        not import_obj.import_statement.asname
+                        isinstance(import_obj, Import) and
+                        not import_obj.key.asname
                 ):
                     seen_module_names.update(
-                        _module_to_base_modules(
-                            import_obj.import_statement.module,
-                        ),
+                        _module_to_base_modules(import_obj.module),
                     )
                 without_exact_duplicates.append(partition)
         else:
@@ -315,9 +297,9 @@ def remove_duplicated_imports(
         if partition.code_type is CodeType.IMPORT:
             import_obj = import_obj_from_str(partition.src)
             if (
-                    isinstance(import_obj, ImportImport) and
-                    not import_obj.import_statement.asname and
-                    import_obj.import_statement.module in seen_module_names
+                    isinstance(import_obj, Import) and
+                    not import_obj.key.asname and
+                    import_obj.key.module in seen_module_names
             ):
                 continue
         out_partitions.append(partition)
@@ -327,7 +309,7 @@ def remove_duplicated_imports(
 
 def apply_import_sorting(
         partitions: Iterable[CodePartition],
-        **sort_kwargs: Any,
+        settings: Settings = Settings(),
 ) -> list[CodePartition]:
     pre_import_code: list[CodePartition] = []
     imports: list[CodePartition] = []
@@ -356,7 +338,7 @@ def apply_import_sorting(
 
     new_imports = []
 
-    sorted_blocks = sort(import_obj_to_partition.keys(), **sort_kwargs)
+    sorted_blocks = sort(import_obj_to_partition.keys(), settings=settings)
     for block in sorted_blocks:
         for import_obj in block:
             new_imports.append(import_obj_to_partition[import_obj])
@@ -368,54 +350,53 @@ def apply_import_sorting(
     if new_imports:
         new_imports.pop()
 
-    # There's the potential that we moved a bunch of whitespace onto the
-    # beginning of the rest of the code.  To fix this, we're going to combine
-    # all of that code, and then make sure there are two linebreaks to start
-    restsrc = _partitions_to_src(rest)
-    restsrc = restsrc.rstrip()
-    if restsrc:
-        rest = [CodePartition(CodeType.CODE, restsrc + '\n')]
-    else:
-        rest = []
-
     return pre_import_code + new_imports + rest
 
 
-def _most_common_line_ending(s: str) -> str:
-    # initialize in case there's no line endings at all
-    counts = collections.Counter({'\n': 0})
-    for line in s.splitlines(True):
-        for ending in ('\r\n', '\r', '\n'):
-            if line.endswith(ending):
-                counts[ending] += 1
-                break
-    return counts.most_common(1)[0][0]
+def _first_line_ending(s: str) -> str:
+    sio = io.StringIO(s, newline='')
+    line = sio.readline()
+    for nl in ('\r\n', '\r'):
+        if line.endswith(nl):
+            return nl
+    else:
+        return '\n'
 
 
 def fix_file_contents(
         contents: str,
-        imports_to_add: tuple[str, ...] = (),
-        imports_to_remove: tuple[str, ...] = (),
-        imports_to_replace: Iterable[ImportToReplace] = (),
-        **sort_kwargs: Any,
+        *,
+        to_add: tuple[str, ...] = (),
+        to_remove: set[tuple[str, ...]],
+        to_replace: Replacements,
+        settings: Settings = Settings(),
 ) -> str:
     # internally use `'\n` as the newline and normalize at the very end
-    nl = _most_common_line_ending(contents)
-    contents = contents.replace('\r\n', '\n').replace('\r', '\n')
+    nl = _first_line_ending(contents)
+    contents = contents.replace('\r\n', '\n').replace('\r', '\n').rstrip()
+    if contents:
+        contents += '\n'
+    else:
+        return ''
 
     partitioned = partition_source(contents)
-    partitioned = combine_trailing_code_chunks(partitioned)
-    partitioned = add_imports(partitioned, to_add=imports_to_add)
+    partitioned = add_imports(partitioned, to_add=to_add)
     partitioned = separate_comma_imports(partitioned)
-    partitioned = remove_imports(partitioned, to_remove=imports_to_remove)
-    partitioned = replace_imports(partitioned, to_replace=imports_to_replace)
-    partitioned = remove_duplicated_imports(partitioned)
-    partitioned = apply_import_sorting(partitioned, **sort_kwargs)
+    partitioned = replace_imports(partitioned, to_replace=to_replace)
+    partitioned = remove_duplicated_imports(partitioned, to_remove=to_remove)
+    partitioned = apply_import_sorting(partitioned, settings=settings)
 
     return _partitions_to_src(partitioned).replace('\n', nl)
 
 
-def _fix_file(filename: str, args: argparse.Namespace) -> int:
+def _fix_file(
+        filename: str,
+        args: argparse.Namespace,
+        *,
+        to_remove: set[tuple[str, ...]],
+        to_replace: Replacements,
+        settings: Settings = Settings(),
+) -> int:
     if filename == '-':
         contents_bytes = sys.stdin.buffer.read()
     else:
@@ -432,11 +413,10 @@ def _fix_file(filename: str, args: argparse.Namespace) -> int:
 
     new_contents = fix_file_contents(
         contents,
-        imports_to_add=args.add_import,
-        imports_to_remove=args.remove_import,
-        imports_to_replace=args.replace_import,
-        application_directories=args.application_directories.split(':'),
-        unclassifiable_application_modules=args.unclassifiable,
+        to_add=args.add_import,
+        to_remove=to_remove,
+        to_replace=to_replace,
+        settings=settings,
     )
     if filename == '-':
         print(new_contents, end='')
@@ -697,6 +677,36 @@ REPLACES[(3,)].update((
 ))
 # END GENERATED
 
+# GENERATED VIA generate-collections-info
+REPLACES[(3,)].update((
+    'collections=collections.abc:AsyncGenerator',
+    'collections=collections.abc:AsyncIterable',
+    'collections=collections.abc:AsyncIterator',
+    'collections=collections.abc:Awaitable',
+    'collections=collections.abc:ByteString',
+    'collections=collections.abc:Callable',
+    'collections=collections.abc:Collection',
+    'collections=collections.abc:Container',
+    'collections=collections.abc:Coroutine',
+    'collections=collections.abc:Generator',
+    'collections=collections.abc:Hashable',
+    'collections=collections.abc:ItemsView',
+    'collections=collections.abc:Iterable',
+    'collections=collections.abc:Iterator',
+    'collections=collections.abc:KeysView',
+    'collections=collections.abc:Mapping',
+    'collections=collections.abc:MappingView',
+    'collections=collections.abc:MutableMapping',
+    'collections=collections.abc:MutableSequence',
+    'collections=collections.abc:MutableSet',
+    'collections=collections.abc:Reversible',
+    'collections=collections.abc:Sequence',
+    'collections=collections.abc:Set',
+    'collections=collections.abc:Sized',
+    'collections=collections.abc:ValuesView',
+))
+# END GENERATED
+
 
 def _add_version_options(parser: argparse.ArgumentParser) -> None:
     versions = sorted(REMOVALS.keys() | REPLACES.keys())
@@ -723,7 +733,7 @@ def _validate_import(s: str) -> str:
         return s
 
 
-def _validate_replace_import(s: str) -> ImportToReplace:
+def _validate_replace_import(s: str) -> tuple[str, str, str]:
     mods, _, attr = s.partition(':')
     try:
         orig_mod, new_mod = mods.split('=')
@@ -732,7 +742,7 @@ def _validate_replace_import(s: str) -> ImportToReplace:
             f'expected `orig.mod=new.mod` or `orig.mod=new.mod:attr`: {s!r}',
         )
     else:
-        return orig_mod.split('.'), new_mod.split('.'), attr
+        return orig_mod, new_mod, attr
 
 
 def main(argv: Sequence[str] | None = None) -> int:
@@ -786,9 +796,17 @@ def main(argv: Sequence[str] | None = None) -> int:
 
     args = parser.parse_args(argv)
 
-    for k, v in REMOVALS.items():
-        if args.min_version >= k:
-            args.remove_import.extend(v)
+    to_remove = {
+        obj.key
+        for s in args.remove_import
+        for obj in import_obj_from_str(s).split()
+    } | {
+        obj.key
+        for k, v in REMOVALS.items()
+        if args.min_version >= k
+        for s in v
+        for obj in import_obj_from_str(s).split()
+    }
 
     for k, v in REPLACES.items():
         if args.min_version >= k:
@@ -796,13 +814,26 @@ def main(argv: Sequence[str] | None = None) -> int:
                 _validate_replace_import(replace_s) for replace_s in v
             )
 
+    to_replace = Replacements.make(args.replace_import)
+
     if os.environ.get('PYTHONPATH'):
         sys.stderr.write('$PYTHONPATH set, import order may be unexpected\n')
         sys.stderr.flush()
 
+    settings = Settings(
+        application_directories=tuple(args.application_directories.split(':')),
+        unclassifiable_application_modules=frozenset(args.unclassifiable),
+    )
+
     retv = 0
     for filename in args.filenames:
-        retv |= _fix_file(filename, args)
+        retv |= _fix_file(
+            filename,
+            args,
+            to_remove=to_remove,
+            to_replace=to_replace,
+            settings=settings,
+        )
     return retv
 
 
